/**
 * @copyright Cube Dev, Inc.
 * @license Apache-2.0
 * @fileoverview The `BaseDriver` and related types declaration.
 */
/// <reference types="node" />
/// <reference types="node" />
import * as stream from 'stream';
import type { ConnectionOptions as TLSConnectionOptions } from 'tls';
import { S3ClientConfig } from '@aws-sdk/client-s3';
import { ExternalCreateTableOptions, DownloadQueryResultsOptions, DownloadQueryResultsResult, DownloadTableData, DriverInterface, ExternalDriverCompatibilities, IndexesSQL, QueryOptions, TableColumn, TableQueryResult, TableStructure, DriverCapabilities, QuerySchemasResult, QueryTablesResult, QueryColumnsResult, TableMemoryData, PrimaryKeysQueryResult, ForeignKeysQueryResult } from './driver.interface';
/**
 * @see {@link DefaultAzureCredential} constructor options
 */
export type AzureStorageClientConfig = {
    azureKey?: string;
    sasToken?: string;
    /**
     * The client ID of a Microsoft Entra app registration.
     * In case of DefaultAzureCredential flow if it is omitted
     * the Azure library will try to use the AZURE_CLIENT_ID env
     */
    clientId?: string;
    /**
     * ID of the application's Microsoft Entra tenant. Also called its directory ID.
     * In case of DefaultAzureCredential flow if it is omitted
     * the Azure library will try to use the AZURE_TENANT_ID env
     */
    tenantId?: string;
    /**
     * The path to a file containing a Kubernetes service account token that authenticates the identity.
     * In case of DefaultAzureCredential flow if it is omitted
     * the Azure library will try to use the AZURE_FEDERATED_TOKEN_FILE env
     */
    tokenFilePath?: string;
};
export type GoogleStorageClientConfig = {
    credentials: any;
};
/**
 * Base driver class.
 */
export declare abstract class BaseDriver implements DriverInterface {
    private testConnectionTimeoutValue;
    protected logger: any;
    /**
     * Class constructor.
     */
    constructor(_options?: {
        /**
         * Time to wait for a response from a connection after validation
         * request before determining it as not valid. Default - 10000 ms.
         */
        testConnectionTimeout?: number;
    });
    protected informationSchemaQuery(): string;
    protected getSchemasQuery(): string;
    protected getTablesForSpecificSchemasQuery(schemasPlaceholders: string): string;
    protected getColumnsForSpecificTablesQuery(conditionString: string): string;
    protected primaryKeysQuery(_?: string): string | null;
    protected foreignKeysQuery(_?: string): string | null;
    protected primaryKeys(conditionString?: string, params?: string[]): Promise<PrimaryKeysQueryResult[]>;
    protected foreignKeys(conditionString?: string, params?: string[]): Promise<ForeignKeysQueryResult[]>;
    protected getColumnNameForSchemaName(): string;
    protected getColumnNameForTableName(): string;
    protected getSslOptions(dataSource: string): TLSConnectionOptions | undefined;
    abstract testConnection(): Promise<void>;
    abstract query<R = unknown>(_query: string, _values?: unknown[], _options?: QueryOptions): Promise<R[]>;
    streamQuery(sql: string, values: string[]): Promise<stream.Readable>;
    downloadQueryResults(query: string, values: unknown[], _options: DownloadQueryResultsOptions): Promise<DownloadQueryResultsResult>;
    readOnly(): boolean;
    protected informationColumnsSchemaReducer(result: any, i: any): any;
    tablesSchema(): Promise<any>;
    tablesSchemaV2(): Promise<any>;
    createSchemaIfNotExists(schemaName: string): Promise<void>;
    getSchemas(): Promise<QuerySchemasResult[]>;
    getTablesForSpecificSchemas(schemas: QuerySchemasResult[]): Promise<QueryTablesResult[]>;
    getColumnsForSpecificTables(tables: QueryTablesResult[]): Promise<QueryColumnsResult[]>;
    getTablesQuery(schemaName: string): Promise<TableQueryResult[]>;
    loadPreAggregationIntoTable(_preAggregationTableName: string, loadSql: string, params: any, options: any): Promise<unknown[]>;
    dropTable(tableName: string, options?: QueryOptions): Promise<unknown>;
    param(_paramIndex: number): string;
    testConnectionTimeout(): number;
    downloadTable(table: string, _options: ExternalDriverCompatibilities): Promise<TableMemoryData>;
    uploadTable(table: string, columns: TableStructure, tableData: DownloadTableData): Promise<void>;
    uploadTableWithIndexes(table: string, columns: TableStructure, tableData: DownloadTableData, indexesSql: IndexesSQL, _uniqueKeyColumns: string[] | null, _queryTracingObj: any, _externalOptions: ExternalCreateTableOptions): Promise<void>;
    protected toColumnValue(value: string, _genericType: string): string | boolean;
    tableColumnTypes(table: string): Promise<TableStructure>;
    queryColumnTypes(sql: string, params: unknown[]): Promise<{
        name: any;
        type: string;
    }[]>;
    createTable(quotedTableName: string, columns: TableColumn[]): Promise<unknown[]>;
    protected createTableSql(quotedTableName: string, columns: TableColumn[]): string;
    protected toGenericType(columnType: string): string;
    protected fromGenericType(columnType: string): string;
    protected quoteIdentifier(identifier: string): string;
    protected cancelCombinator(fn: any): any;
    setLogger(logger: any): void;
    protected reportQueryUsage(usage: any, queryOptions: any): void;
    protected databasePoolError(error: any): void;
    release(): Promise<void>;
    capabilities(): DriverCapabilities;
    nowTimestamp(): number;
    wrapQueryWithLimit(query: {
        query: string;
        limit: number;
    }): void;
    /**
     * Returns an array of signed AWS S3 URLs of the unloaded csv files.
     */
    protected extractUnloadedFilesFromS3(clientOptions: S3ClientConfig, bucketName: string, prefix: string): Promise<string[]>;
    /**
     * Returns an array of signed GCS URLs of the unloaded csv files.
     */
    protected extractFilesFromGCS(gcsConfig: GoogleStorageClientConfig, bucketName: string, tableName: string): Promise<string[]>;
    protected extractFilesFromAzure(azureConfig: AzureStorageClientConfig, bucketName: string, tableName: string): Promise<string[]>;
}
//# sourceMappingURL=BaseDriver.d.ts.map